package chess4j.io;

import java.text.ParseException;
import java.util.List;
import java.util.Timer;

import groovy.lang.MissingMethodException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import chess4j.App;
import chess4j.Color;
import chess4j.board.Board;
import chess4j.board.File;
import chess4j.board.Rank;
import chess4j.board.Square;
import chess4j.eval.Eval;
import chess4j.exceptions.IllegalMoveException;
import chess4j.moves.Move;
import chess4j.search.Search;
import chess4j.search.SearchIterator;
import chess4j.utils.GameStatus;
import chess4j.utils.GameStatusChecker;
import chess4j.utils.Perft;



public class InputParser {
	private static final Log logger = LogFactory.getLog(InputParser.class);
	private static final InputParser INSTANCE = new InputParser();
	private boolean forceMode;
	private Thread searchThread;
	private boolean icsMode=false;
	
	private InputParser() { }
	
	public static InputParser getInstance() {
		return INSTANCE;
	}
	
	public void parseCommand(String command) throws IllegalMoveException,MissingMethodException,ParseException {
		def input = command.tokenize()
		def method = input.remove(0)
		//logger.info ("Method: " + method)
		//logger.info ("Input: " + input)
		if ("new".equals(method)) {  // because 'new' is a reserved word
			invokeMethod("newGame",input)
		} else if ("?".equals(method)) { // can't have a method named '?'
			invokeMethod("moveNow",input)
		} else {
			invokeMethod(method,input)
		}
	}
	
	def accepted(def arg) {}
	
	def analyze(def arg) {}

	def black(def arg) {}
	
	def bk(def arg) {}
	
	def computer(def arg) {}

	def db(def arg) {
		joinSearchThread()
		DrawBoard.drawBoard(Board.INSTANCE);
	}

	def draw(def arg) {}
	
	def easy(def arg) {}

	def eval(def arg) {
		joinSearchThread()
		int eval = Eval.eval(Board.INSTANCE);
		logger.info("eval=" + eval)
	}

	def force(def arg) {
		forceMode = true;
	}
		
	def go(def arg) {
		joinSearchThread();
		forceMode = false;
		thinkAndMakeMove();
	}
	
	def hard(def arg) {}
	
	def hint(def arg) {}
	
	def ics(def arg) {
		synchronized(this) {
			if (icsMode==false) {
				logger.info "# ICS mode on."
				java.util.TimerTask task = new java.util.TimerTask() {
					public void run() {
						logger.info "tellics open\n"
						logger.info "tellics gameend\n"
					}
				};
				Timer timer = new Timer()
				timer.scheduleAtFixedRate(task,0,30*60*1000); // 30 minutes
				icsMode = true;
			}
		}
	}
	
	def level(List args) {
		String mps = args.get(0)
		String base = args.get(1)
		Integer increment = Integer.valueOf(args.get(2)) 
		logger.debug "# level: " + mps + ", " + base + ", " + increment
		increment *= 1000
		logger.debug "# setting increment to " + increment + " ms."
		SearchIterator.setIncrement(increment)
	}

	def moveNow(def arg) {
		Search.setAbortSearch(true)
		joinSearchThread()	
	}

	def name(List args) {
		String opponent = args.get(0)
		logger.info "# opponent is: " + opponent
	}
	
	def newGame(def arg) {
		joinSearchThread();
		forceMode = false;
		Board.INSTANCE.resetBoard();
		SearchIterator.setMaxDepth(0)
	}

	def nopost(def arg) { }
	
	def otim(def arg) { }

	def perft(List args) {
		joinSearchThread();
		int depth = Integer.valueOf(args.get(0));
		DrawBoard.drawBoard(Board.INSTANCE);
		long start = System.currentTimeMillis();
		long nodes=Perft.perft(Board.INSTANCE, depth);
		long end = System.currentTimeMillis();
		if (end==start) end=start+1; // HACK to avoid div 0
		logger.info "# nodes: " + nodes
		logger.info "# elapsed time: " + (end-start) + " ms"
		logger.info "# rate: " + nodes*1000/(end-start) + " n/s\n"
	}

	/**
	* Leave force mode and set the engine to play the color that is not on move.
	* Associate the opponent's clock with the color that is on move, the engine's
	* clock with the color that is not on move.  Start the opponent's clock.  If
	* pondering is enabled, the engine should begin pondering.  If the engine later
	* receives a move, it should start thinking and eventually reply.
	*/
	def playother(def arg) { 
		joinSearchThread()
		forceMode = false
		// TODO: OK to start pondering here
	}
	
	/**
	 * Because of the way xboard uses the ping command, we should never see a "ping" command
	 * when the engine is on move.  However, if we do, we should not respond with the "pong"
	 * until after making the move.  The exception is pondering -- if the engine is pondering
	 * it should immediately respond with the pong but continue pondering.
	 *
	 * The documentation specifically mentions a "?" followed by "ping".  In this case
	 * the pong should not be sent until after the move is made.
	 *
	 */
	def ping(List args) {
		joinSearchThread() // TODO: this might change with pondering
		logger.info "pong " + args.get(0)
	}	
	
	def post(def arg) { }

	def protover(List args) {
		Integer version = Integer.valueOf(args.get(0))
		if (version < 2) {
			logger.info "Error: invalid protocol version."
			System.exit(1)
		}
		logger.info "feature analyze=0 black=0 colors=0 ping=1 draw=0 debug=1 edit=0 ics=1"
		logger.info "feature level=0 name=1 nps=0 memory=0 playother=1 pause=0 resume=0 reuse=1 san=0"
		logger.info "feature setboard=1 sigint=0 sigterm=0 smp=0 st=0 time=1 usermove=1"
		logger.info "white=0 variants=\"normal\""
		logger.info "feature done=1" // must be last
	}

	def quit(def arg) {
		Search.setAbortSearch(true)
		joinSearchThread()
		logger.info "bye..."
		System.exit(0);
	}

	def random(def arg) { }
	
	def rating(def arg) { }
	
	def rejected(def arg) { }
	
	/**
	* Retract a move.  Undoes the last two moves and continues playing the same color.
	* Xboard sends this command only when the user is on move.
	*/
	def remove(def arg) { 
		joinSearchThread()
		Board.INSTANCE.undoLastMove()
		Board.INSTANCE.undoLastMove()
	}

	/**
	 * RESULT is either 1-0, 0-1, 1/2-1/2, or *, indicating whether white won, black won,
	 * the game was drawn, or the game unfinished.  The COMMENT string is purely human
	 * readable.  It is subject to change.
	 *
	 * We will get a RESULT command even if we already know the game ended.  If we send
	 * the RESULT command (e.g. after mating the opponent), we will still get a RESULT
	 * command back.
	 *
	 * We will NOT get a RESULT command if the user stops playing by selecting Reset,
	 * Edit Game, Exit, or the like.
	 */
	def result(def arg) {
		joinSearchThread()
	}	

	
	/**
	 * sd DEPTH
	 * The engine will limit its thinking to DEPTH ply.  Note the engine will still respect
	 * any time control restrictions in place.
	 *
	 */
	def sd(List args) { 
		Integer depth = Integer.valueOf(args.get(0))
		logger.info "# setting depth to : " + depth
		SearchIterator.setMaxDepth(depth)
	}
			
	def setboard(List args) {
		String fen = args.join(' ')
		FenParser.setPos(Board.INSTANCE,fen);
		DrawBoard.drawBoard(Board.INSTANCE)
	}

	def st(def arg) { }

	/**
	 * Read in the engine's remaining time, in centiseconds.
	 */
	def time(List args) { 
		Integer time = Integer.valueOf(args.get(0))
		logger.info "#time : " + time
		SearchIterator.setTimeRemaining(time)
	}

	/**
	* Back up one move.  Xboard will never send this without putting the engine in
	* "force" mode first.  We don't have to worry about undoing a move the engine made.
	*/
	def undo(def arg) {
		joinSearchThread()
		Board.INSTANCE.undoLastMove()
	}

	def usermove(List args) throws IllegalMoveException {
		joinSearchThread();
		String strMove = args.get(0)
		Move mv = MoveParser.parseLegalMove(strMove,Board.INSTANCE);
		Board.INSTANCE.applyMove(mv);
		if (!forceMode) {
			thinkAndMakeMove();
		}
	}

	def variant(List args) {
		String v = args.get(0)
		if (! "normal".equals(v)) {
			logger.info "Error: unsupported variant."
		}
	}
			
	def white(def arg) {}
	
	def xboard(def arg) {}
	
	private void joinSearchThread() {
		if (searchThread==null) return;
		searchThread.join();
	}
	
	private void thinkAndMakeMove() {
		GameStatus gs = GameStatusChecker.getGameStatus(Board.INSTANCE);
		if (gs != GameStatus.INPROGRESS) {
			PrintGameResult.printResult(gs);
			return;
		}
		Search.setAbortSearch(false);
		searchThread = new Thread(SearchIterator.getInstance());
		searchThread.start();
	}
}
