package chess4j.board;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import chess4j.Color;
import chess4j.attack.Attack;
import chess4j.moves.Move;
import chess4j.pieces.Bishop;
import chess4j.pieces.King;
import chess4j.pieces.Knight;
import chess4j.pieces.Pawn;
import chess4j.pieces.Piece;
import chess4j.pieces.Queen;
import chess4j.pieces.Rook;

public class Board {
	
	public static final Board INSTANCE = new Board();
	
	private List<Undo> undoStack = new ArrayList<Undo>();
	private Map<Square,Piece> pieceMap = new HashMap<Square,Piece>();
	private Set<CastlingRights> castlingRights = new HashSet<CastlingRights>();
	private Color playerToMove;
	private Square epSquare;
	private int moveCounter;
	private int fiftyCounter;
	
	private Board() {
		resetBoard();
	}
		
	public void addCastlingRight(CastlingRights castlingRight) {
		castlingRights.add(castlingRight);
	}

	public void addPiece(Piece p,Square s) {
		if (p==null) return;
		pieceMap.put(s, p);
	}
	
	private void applyKingSpecialCases(Move m) {
		if (playerToMove.isWhite()) {
			castlingRights.remove(CastlingRights.WHITE_KINGSIDE);
			castlingRights.remove(CastlingRights.WHITE_QUEENSIDE);
			if (m.from().equals(Square.valueOf(File.FILE_E, Rank.RANK_1))) {
				if (m.to().equals(Square.valueOf(File.FILE_G, Rank.RANK_1))) {
					fiftyCounter=0;
					movePiece(Square.valueOf(File.FILE_H, Rank.RANK_1),
							Square.valueOf(File.FILE_F,Rank.RANK_1));
				} else if (m.to().equals(Square.valueOf(File.FILE_C, Rank.RANK_1))) {
					fiftyCounter=0;
					movePiece(Square.valueOf(File.FILE_A, Rank.RANK_1),
							Square.valueOf(File.FILE_D,Rank.RANK_1));
				}
			}
		} else {
			castlingRights.remove(CastlingRights.BLACK_KINGSIDE);
			castlingRights.remove(CastlingRights.BLACK_QUEENSIDE);
			if (m.from().equals(Square.valueOf(File.FILE_E, Rank.RANK_8))) {
				if (m.to().equals(Square.valueOf(File.FILE_G, Rank.RANK_8))) {
					fiftyCounter=0;
					movePiece(Square.valueOf(File.FILE_H, Rank.RANK_8),
							Square.valueOf(File.FILE_F,Rank.RANK_8));
				} else if (m.to().equals(Square.valueOf(File.FILE_C, Rank.RANK_8))) {
					fiftyCounter=0;
					movePiece(Square.valueOf(File.FILE_A, Rank.RANK_8),
							Square.valueOf(File.FILE_D,Rank.RANK_8));
				}
			}
		}
	}
	
	public void applyMove(Move m) {
		undoStack.add(new Undo(m,fiftyCounter,castlingRights,epSquare));

		moveCounter++;
		if (m.captured()==null) {
			fiftyCounter++; // may still get set to 0
		} else {
			fiftyCounter=0;
			if (m.to().equals(Square.valueOf(File.FILE_H, Rank.RANK_1))) {
				castlingRights.remove(CastlingRights.WHITE_KINGSIDE);
			} else if (m.to().equals(Square.valueOf(File.FILE_A, Rank.RANK_1))) {
				castlingRights.remove(CastlingRights.WHITE_QUEENSIDE);
			} if (m.to().equals(Square.valueOf(File.FILE_H, Rank.RANK_8))) {
				castlingRights.remove(CastlingRights.BLACK_KINGSIDE);
			} else if (m.to().equals(Square.valueOf(File.FILE_A, Rank.RANK_8))) {
				castlingRights.remove(CastlingRights.BLACK_QUEENSIDE);
			}
		}

		Piece p = removePiece(m.from());
		if (p instanceof Pawn) {
			applyPawnMove(p,m);
		} else {
			if (m.captured() != null) {
				removePiece(m.to());
			}
			addPiece(p,m.to());
			epSquare=null;
			if (p instanceof King) 
				applyKingSpecialCases(m);
			else if (p instanceof Rook)
				applyRookSpecialCases(m);
		}
		swapPlayer();
	}

	private void applyPawnMove(Piece p,Move m) {
		fiftyCounter=0;
		if (p.isWhite()) {
			if (m.from().rank().equals(Rank.RANK_2) && m.to().rank().equals(Rank.RANK_4)) {
				addPiece(p,m.to());
				epSquare = Square.valueOf(m.to().file(), m.to().rank().south());
			} else if (m.to().equals(epSquare)) {
				removePiece(Square.valueOf(epSquare.file(), epSquare.rank().south()));
				addPiece(p,m.to());
				epSquare = null;
			} else if (m.to().rank().equals(Rank.RANK_8)) {
				removePiece(m.to());
				addPiece(m.promotion(),m.to());
				epSquare = null;
			} else {
				removePiece(m.to());
				addPiece(p,m.to());
				epSquare = null;
			}
		} else {
			if (m.from().rank().equals(Rank.RANK_7) && m.to().rank().equals(Rank.RANK_5)) {
				addPiece(p,m.to());
				epSquare = Square.valueOf(m.to().file(), m.to().rank().north());
			} else if (m.to().equals(epSquare)) {
				removePiece(Square.valueOf(epSquare.file(), epSquare.rank().north()));
				addPiece(p,m.to());
				epSquare = null;
			} else if (m.to().rank().equals(Rank.RANK_1)) {
				removePiece(m.to());
				addPiece(m.promotion(),m.to());
				epSquare = null;
			} else {
				removePiece(m.to());
				addPiece(p,m.to());
				epSquare = null;
			}
		}
	}

	private void applyRookSpecialCases(Move m) {
		if (playerToMove.isWhite()) {
			if (m.from().equals(Square.valueOf(File.FILE_H, Rank.RANK_1))) {
				fiftyCounter=0;
				castlingRights.remove(CastlingRights.WHITE_KINGSIDE);
			} else if (m.from().equals(Square.valueOf(File.FILE_A, Rank.RANK_1))) {
				fiftyCounter=0;
				castlingRights.remove(CastlingRights.WHITE_QUEENSIDE);
			}
		} else {
			if (m.from().equals(Square.valueOf(File.FILE_H, Rank.RANK_8))) {
				fiftyCounter=0;
				castlingRights.remove(CastlingRights.BLACK_KINGSIDE);
			} else if (m.from().equals(Square.valueOf(File.FILE_A, Rank.RANK_8))) {
				fiftyCounter=0;
				castlingRights.remove(CastlingRights.BLACK_QUEENSIDE);
			}			
		}
	}

	private boolean blackCanCastleKingSide() {
		Color opponent = Color.swap(playerToMove);
		return hasCastlingRight(CastlingRights.BLACK_KINGSIDE)
			&& isEmpty(Square.valueOf(File.FILE_F,Rank.RANK_8)) 
			&& isEmpty(Square.valueOf(File.FILE_G,Rank.RANK_8))
			&& !Attack.attacked(this,Square.valueOf(File.FILE_E,Rank.RANK_8),opponent)
			&& !Attack.attacked(this,Square.valueOf(File.FILE_F,Rank.RANK_8),opponent)
			;
	}

	private boolean blackCanCastleQueenSide() {
		Color opponent = Color.swap(playerToMove);
		return hasCastlingRight(CastlingRights.BLACK_QUEENSIDE)
			&& isEmpty(Square.valueOf(File.FILE_D,Rank.RANK_8)) 
			&& isEmpty(Square.valueOf(File.FILE_C,Rank.RANK_8)) 
			&& isEmpty(Square.valueOf(File.FILE_B,Rank.RANK_8))
			&& !Attack.attacked(this,Square.valueOf(File.FILE_E,Rank.RANK_8),opponent)
			&& !Attack.attacked(this,Square.valueOf(File.FILE_D,Rank.RANK_8),opponent)
			;
	}
	
	public boolean canCastle(CastlingRights cr) {
		if (cr.equals(CastlingRights.WHITE_KINGSIDE))
			return whiteCanCastleKingSide();
		else if (cr.equals(CastlingRights.WHITE_QUEENSIDE))
			return whiteCanCastleQueenSide();
		else if (cr.equals(CastlingRights.BLACK_KINGSIDE))
			return blackCanCastleKingSide();
		else return blackCanCastleQueenSide();
	}
	
	public void clearBoard() {
		Set<Square> squares = Square.allSquares();
		for (Square sq : squares) {
			Piece p = getPiece(sq);
			if (p != null) {
				removePiece(sq);
			}
		}
		
		epSquare = null;
		castlingRights.clear();
		fiftyCounter = 0;
		undoStack.clear();
	}

	public synchronized Board deepCopy() {
		Board b = new Board();
		b.undoStack.clear();
		b.undoStack.addAll(undoStack);
		b.pieceMap.clear();
		for (Square sq : pieceMap.keySet()) {
			b.pieceMap.put(sq, pieceMap.get(sq));
		}
		b.castlingRights.clear();
		b.castlingRights.addAll(castlingRights);
		b.playerToMove=playerToMove;
		b.epSquare=epSquare;
		b.fiftyCounter=fiftyCounter;
		b.moveCounter=moveCounter;
		
		return b;
	}
	
	public void flipVertical() {
		
		Set<Square> squares = Square.allSquares();
		Map<Square,Piece> myPieceMap = new HashMap<Square,Piece>(); 
		// remove pieces, remembering where they were
		for (Square sq : squares) {
			Piece p = removePiece(sq);
			myPieceMap.put(sq, p);
		}
		
		// now flip them around
		for (Square sq : squares) {
			Piece p = myPieceMap.get(sq);
			if (p != null) {
				addPiece(p.getOppositeColorPiece(),sq.flipVertical());
			}
		}
		
		swapPlayer();
		
		if (epSquare != null) {
			epSquare = epSquare.flipVertical();
		}
		
		// flip castling rights
		Set<CastlingRights> myCastlingRights = new HashSet<CastlingRights>();
		myCastlingRights.addAll(castlingRights);
		castlingRights.clear();
		for (CastlingRights cr : myCastlingRights) {
			if (CastlingRights.WHITE_KINGSIDE.equals(cr)) {
				castlingRights.add(CastlingRights.BLACK_KINGSIDE);
			} else if (CastlingRights.WHITE_QUEENSIDE.equals(cr)) {
				castlingRights.add(CastlingRights.BLACK_QUEENSIDE);
			} else if (CastlingRights.BLACK_KINGSIDE.equals(cr)) {
				castlingRights.add(CastlingRights.WHITE_KINGSIDE);
			} else if (CastlingRights.BLACK_QUEENSIDE.equals(cr)) { 
				castlingRights.add(CastlingRights.WHITE_QUEENSIDE);
			}
		}
		
	}
	
	@Override
	public boolean equals(Object obj) {
		if (!(obj instanceof Board)) return false;
		Board that = (Board)obj;
		if (!equalExceptMoveHistory(that,true)) return false;
		if (!this.undoStack.equals(that.undoStack)) return false;
		return true;
	}
	
	public boolean equalExceptMoveHistory(Board otherBoard,boolean strict) {
		if (!this.pieceMap.equals(otherBoard.pieceMap)) return false;
		if (!this.castlingRights.equals(otherBoard.castlingRights)) return false;
		if (!this.getPlayerToMove().equals(otherBoard.getPlayerToMove())) return false;
		if (this.epSquare==null) {
			if (otherBoard.epSquare!=null) return false;
		} else {
			if (!this.epSquare.equals(otherBoard.epSquare)) return false;
		}
		if (strict) {
			if (this.moveCounter!=otherBoard.moveCounter) return false;
			if (this.fiftyCounter!=otherBoard.fiftyCounter) return false;
		}
		return true;
	}
	
	private Square getBlackKingSquare() {
		Set<Square> squares = Square.allSquares();
		for (Square sq : squares) {
			Piece p = getPiece(sq);
			if (King.BLACK_KING.equals(p)) {
				return sq;
			}
		}
		return null;
	}
	
	public Square getEPSquare() {
		return epSquare;
	}
	
	public int getFiftyCounter() {
		return fiftyCounter;
	}
	
	public Square getKingSquare(Color player) {
		return player.isWhite() ? getWhiteKingSquare() : getBlackKingSquare();
	}

	public int getMoveCounter() {
		return moveCounter;
	}
	
	public Piece getPiece(Square square) {
		return pieceMap.get(square);
	}
	
	public Color getPlayerToMove() {
		return playerToMove;
	}

	/**
	 * TODO
	 * @return
	 */
	private Square getWhiteKingSquare() {
		Set<Square> squares = Square.allSquares();
		for (Square sq : squares) {
			Piece p = getPiece(sq);
			if (King.WHITE_KING.equals(p)) {
				return sq;
			}
		}
		return null;
	}
	
	public boolean hasCastlingRight(CastlingRights cr) {
		return castlingRights.contains(cr);
	}

	public boolean isEmpty(Square square) {
		return pieceMap.get(square)==null;
	}

	public boolean isOpponentInCheck() {
		return Attack.attacked(this,getKingSquare(Color.swap(playerToMove)),playerToMove);
	}

	public boolean isPlayerInCheck() {
		return Attack.attacked(this,getKingSquare(playerToMove),Color.swap(playerToMove));
	}

	public void movePiece(Square from,Square to) {
		if (from.equals(to))
			return;
		addPiece(getPiece(from),to);
		removePiece(from);
	}

	private Piece removePiece(Square s) {
		Piece p = getPiece(s);
		if (p==null) return null;
		pieceMap.remove(s);
		return p;
	}
	
	public void resetBoard() {
		undoStack.clear();
		pieceMap.clear();
		addPiece(Rook.BLACK_ROOK,Square.valueOf(File.FILE_A, Rank.RANK_8));
		addPiece(Knight.BLACK_KNIGHT,Square.valueOf(File.FILE_B, Rank.RANK_8));
		addPiece(Bishop.BLACK_BISHOP,Square.valueOf(File.FILE_C, Rank.RANK_8));
		addPiece(Queen.BLACK_QUEEN,Square.valueOf(File.FILE_D, Rank.RANK_8));
		addPiece(King.BLACK_KING,Square.valueOf(File.FILE_E, Rank.RANK_8));
		addPiece(Bishop.BLACK_BISHOP,Square.valueOf(File.FILE_F, Rank.RANK_8));
		addPiece(Knight.BLACK_KNIGHT,Square.valueOf(File.FILE_G, Rank.RANK_8));
		addPiece(Rook.BLACK_ROOK,Square.valueOf(File.FILE_H, Rank.RANK_8));
		Set<Square> squares = Square.rankSquares(Rank.RANK_7);
		for (Square sq : squares) {
			addPiece(Pawn.BLACK_PAWN,sq);
		}
		squares = Square.rankSquares(Rank.RANK_2);
		for (Square sq : squares) {
			addPiece(Pawn.WHITE_PAWN,sq);
		}
		addPiece(Rook.WHITE_ROOK,Square.valueOf(File.FILE_A, Rank.RANK_1));
		addPiece(Knight.WHITE_KNIGHT,Square.valueOf(File.FILE_B, Rank.RANK_1));
		addPiece(Bishop.WHITE_BISHOP,Square.valueOf(File.FILE_C, Rank.RANK_1));
		addPiece(Queen.WHITE_QUEEN,Square.valueOf(File.FILE_D, Rank.RANK_1));
		addPiece(King.WHITE_KING,Square.valueOf(File.FILE_E, Rank.RANK_1));
		addPiece(Bishop.WHITE_BISHOP,Square.valueOf(File.FILE_F, Rank.RANK_1));
		addPiece(Knight.WHITE_KNIGHT,Square.valueOf(File.FILE_G, Rank.RANK_1));
		addPiece(Rook.WHITE_ROOK,Square.valueOf(File.FILE_H, Rank.RANK_1));
	
		castlingRights.clear();
		castlingRights.addAll(EnumSet.allOf(CastlingRights.class));
		playerToMove = Color.WHITE;
		epSquare = null;
		moveCounter = 0;
		fiftyCounter = 0;
	}

	public void setEP(Square ep) {
		epSquare = ep;
	}

	public void setFiftyCounter(int fiftyCounter) {
		this.fiftyCounter=fiftyCounter;
	}
	
	public void setMoveCounter(int moveCounter) {
		this.moveCounter=moveCounter;
	}
	
	public void swapPlayer() {
		playerToMove = Color.swap(playerToMove);
	}

	private void undoCastle(Undo u,Piece p) {
		if (p.isWhite()) {
			if (u.getMove().from().equals(Square.valueOf(File.FILE_E, Rank.RANK_1))) {
				if (u.getMove().to().equals(Square.valueOf(File.FILE_G, Rank.RANK_1))) {
					movePiece(Square.valueOf(File.FILE_F, Rank.RANK_1),
							Square.valueOf(File.FILE_H,Rank.RANK_1));
				} else if (u.getMove().to().equals(Square.valueOf(File.FILE_C, Rank.RANK_1))) {
					movePiece(Square.valueOf(File.FILE_D, Rank.RANK_1),
							Square.valueOf(File.FILE_A,Rank.RANK_1));
				}
			}
		} else {
			if (u.getMove().from().equals(Square.valueOf(File.FILE_E, Rank.RANK_8))) {
				if (u.getMove().to().equals(Square.valueOf(File.FILE_G, Rank.RANK_8))) {
					movePiece(Square.valueOf(File.FILE_F, Rank.RANK_8),
							Square.valueOf(File.FILE_H,Rank.RANK_8));
				} else if (u.getMove().to().equals(Square.valueOf(File.FILE_C, Rank.RANK_8))) {
					movePiece(Square.valueOf(File.FILE_D, Rank.RANK_8),
							Square.valueOf(File.FILE_A,Rank.RANK_8));
				}
			}
		} 
	}
	
	public int getSizeOfUndoStack() {
		return undoStack.size();
	}
	
	public void undoLastMove() {
		if (undoStack.size()==0) return;
		int ind = undoStack.size()-1;
		Undo u = undoStack.remove(ind);
		swapPlayer();
		epSquare = u.getEpSquare();
		moveCounter--;
		fiftyCounter = u.getFiftyCounter();
		castlingRights.clear();
		castlingRights.addAll(u.getCastlingRights());
		
		Piece p = removePiece(u.getMove().to());
		if (u.getMove().promotion() != null) {
			undoPromotion(u);
		} else if (p instanceof Pawn) {
			undoPawnMove(u,p);
		} else { 
			if (p instanceof King) {
				undoCastle(u,p);
			}
			addPiece(u.getMove().captured(),u.getMove().to());
			addPiece(p,u.getMove().from());
		}
	}

	private void undoPawnMove(Undo u,Piece p) {
		if (p.isWhite()) {
			if (u.getMove().to().equals(epSquare)) {
				addPiece(u.getMove().captured(),Square.valueOf(epSquare.file(), epSquare.rank().south()));
				addPiece(p,u.getMove().from());
			} else {
				addPiece(u.getMove().captured(),u.getMove().to());
				addPiece(p,u.getMove().from());
			}
		} else {
			if (u.getMove().to().equals(epSquare)) {
				addPiece(u.getMove().captured(),Square.valueOf(epSquare.file(), epSquare.rank().north()));
				addPiece(p,u.getMove().from());
			} else {
				addPiece(u.getMove().captured(),u.getMove().to());
				addPiece(p,u.getMove().from());
			}
		}			
	}
	
	private void undoPromotion(Undo u) {
		addPiece(u.getMove().captured(),u.getMove().to());
		addPiece(
			playerToMove.equals(Color.WHITE)?Pawn.WHITE_PAWN:Pawn.BLACK_PAWN,
			u.getMove().from());
	}
	
	private boolean whiteCanCastleKingSide() {
		Color opponent = Color.swap(playerToMove);
		return hasCastlingRight(CastlingRights.WHITE_KINGSIDE)
			&& isEmpty(Square.valueOf(File.FILE_F,Rank.RANK_1)) 
			&& isEmpty(Square.valueOf(File.FILE_G,Rank.RANK_1))
			&& !Attack.attacked(this,Square.valueOf(File.FILE_E,Rank.RANK_1),opponent)
			&& !Attack.attacked(this,Square.valueOf(File.FILE_F,Rank.RANK_1),opponent)
			;
	}

	private boolean whiteCanCastleQueenSide() {
		Color opponent = Color.swap(playerToMove);
		return hasCastlingRight(CastlingRights.WHITE_QUEENSIDE)
			&& isEmpty(Square.valueOf(File.FILE_D,Rank.RANK_1)) 
			&& isEmpty(Square.valueOf(File.FILE_C,Rank.RANK_1)) 
			&& isEmpty(Square.valueOf(File.FILE_B,Rank.RANK_1))
			&& !Attack.attacked(this,Square.valueOf(File.FILE_E,Rank.RANK_1),opponent)
			&& !Attack.attacked(this,Square.valueOf(File.FILE_D,Rank.RANK_1),opponent)
			;
	}
	
	/*@Override 
	public int hashCode() {
		int h=0;
		for (Square sq : Square.allSquares()) {
			Piece p = getPiece(sq);
			if (p!=null) {
				h ^= Zobrist.getPieceKey(sq, p);
			}
		}
		h ^= Zobrist.getPlayerKey(getPlayerToMove());
		Square ep = getEPSquare();
		if (ep != null) {
			h ^= Zobrist.getEnPassantKey(ep);
		}
		for (CastlingRights cr : castlingRights) {
			h ^= Zobrist.getCastlingKey(cr);
		}
		return h;
	}*/
}
