package chess4j.moves;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import chess4j.Color;
import chess4j.board.Board;
import chess4j.board.CastlingRights;
import chess4j.board.File;
import chess4j.board.KingSquares;
import chess4j.board.KnightSquares;
import chess4j.board.Rank;
import chess4j.board.Square;
import chess4j.board.directions.Direction;
import chess4j.board.directions.North;
import chess4j.board.directions.NorthEast;
import chess4j.board.directions.NorthWest;
import chess4j.board.directions.PieceDirections;
import chess4j.board.directions.South;
import chess4j.board.directions.SouthEast;
import chess4j.board.directions.SouthWest;
import chess4j.pieces.Bishop;
import chess4j.pieces.King;
import chess4j.pieces.Knight;
import chess4j.pieces.Pawn;
import chess4j.pieces.Piece;
import chess4j.pieces.Queen;
import chess4j.pieces.Rook;


public class MoveGen {

	private MoveGen() {}
	
	private static void addMoves(List<Move> moves,Board board,Square fromSq,Set<Square> toSquares) {
		assert(fromSq != null);
		Piece fromPiece = board.getPiece(fromSq);
		assert(fromPiece != null);
		Color oppColor = Color.swap(fromPiece.getColor()); 
		for (Square toSq : toSquares) {
			Piece toPiece = board.getPiece(toSq);
			if (toPiece==null || toPiece.getColor().equals(oppColor))
				moves.add(new Move(fromSq,toSq,toPiece));
		}
	}

	private static void genBishopMoves(Board board,Square fromSq,List<Move> moves) {
		genSliderMoves(PieceDirections.getBishopDirections(),board,fromSq,moves);
	}

	
	private static void genBlackPawnAttacks(Direction dir,Board board,Square fromSq,List<Move> moves) {
		Square toSq = dir.next(fromSq);
		if (toSq != null) {
			Piece p = board.getPiece(toSq);
			if (p != null && p.getColor().isWhite()) {
				if (toSq.rank().equals(Rank.RANK_1)) {
					moves.add(new Move(fromSq,toSq,p,Queen.BLACK_QUEEN));
					moves.add(new Move(fromSq,toSq,p,Rook.BLACK_ROOK));
					moves.add(new Move(fromSq,toSq,p,Bishop.BLACK_BISHOP));
					moves.add(new Move(fromSq,toSq,p,Knight.BLACK_KNIGHT));
				} else {
					moves.add(new Move(fromSq,toSq,p));
				}
			} else if (toSq.equals(board.getEPSquare())) {
				Square s = Square.valueOf(board.getEPSquare().file(),board.getEPSquare().rank().north());
				moves.add(new Move(fromSq,toSq,board.getPiece(s)));
			}
		}
	}

	private static void genBlackPawnPushes(Board board,Square fromSq,List<Move> moves) {
		Square toSq = South.getInstance().next(fromSq);
		// pawn push
		if (board.isEmpty(toSq)) {
			if (toSq.rank().equals(Rank.RANK_1)) {
				moves.add(new Move(fromSq,toSq,null,Queen.BLACK_QUEEN));
				moves.add(new Move(fromSq,toSq,null,Rook.BLACK_ROOK));
				moves.add(new Move(fromSq,toSq,null,Bishop.BLACK_BISHOP));
				moves.add(new Move(fromSq,toSq,null,Knight.BLACK_KNIGHT));
			} else {
				moves.add(new Move(fromSq,toSq,null));
				// double push
				if (fromSq.rank().equals(Rank.RANK_7)) {
					toSq = South.getInstance().next(toSq);
					if (board.isEmpty(toSq))
						moves.add(new Move(fromSq,toSq,null));
				}
			}
		}		
	}
	
	private static void genCastlingMoves(Board board,List<Move> moves) {
		Color player = board.getPlayerToMove();

		if (player.isWhite()) {
			Square fromSq = Square.valueOf(File.FILE_E, Rank.RANK_1);
			if (board.canCastle(CastlingRights.WHITE_KINGSIDE)) {
				moves.add(new Move(fromSq,Square.valueOf(File.FILE_G, Rank.RANK_1),null));
			}
			if (board.canCastle(CastlingRights.WHITE_QUEENSIDE)) {
				moves.add(new Move(fromSq,Square.valueOf(File.FILE_C,Rank.RANK_1),null));
			}
		} else {
			Square fromSq = Square.valueOf(File.FILE_E, Rank.RANK_8);
			if (board.canCastle(CastlingRights.BLACK_KINGSIDE)) {
				moves.add(new Move(fromSq,Square.valueOf(File.FILE_G, Rank.RANK_8),null));
			}
			if (board.canCastle(CastlingRights.BLACK_QUEENSIDE)) {
				moves.add(new Move(fromSq,Square.valueOf(File.FILE_C,Rank.RANK_8),null));
			}
		}
	}
	
	private static void genKingMoves(Board board,Square fromSq,List<Move> moves) {
		addMoves(moves,board,fromSq,KingSquares.getSquares(fromSq));
		genCastlingMoves(board,moves);
	}
	
	private static void genKnightMoves(Board board,Square fromSq,List<Move> moves) {
		addMoves(moves,board,fromSq,KnightSquares.getSquares(fromSq));
	}

	public static List<Move> genLegalMoves(Board board) {
		List<Move> pseudoLegalMoves = genPseudoLegalMoves(board);
		List<Move> legalMoves = new ArrayList<Move>();
		
		for (Move m : pseudoLegalMoves) {
			board.applyMove(m);
			if (!board.isOpponentInCheck()) {
				legalMoves.add(m);
			}
			board.undoLastMove();
		}

		return legalMoves;
	}

	private static void genPawnMoves(Board board,Square fromSq,List<Move> moves) {
		if (board.getPlayerToMove().isWhite()) {
			genWhitePawnPushes(board,fromSq,moves);
			genWhitePawnAttacks(NorthWest.getInstance(),board,fromSq,moves);
			genWhitePawnAttacks(NorthEast.getInstance(),board,fromSq,moves);
		} else {
			genBlackPawnPushes(board,fromSq,moves);
			genBlackPawnAttacks(SouthWest.getInstance(),board,fromSq,moves);
			genBlackPawnAttacks(SouthEast.getInstance(),board,fromSq,moves);
		}
	}
	
	public static List<Move> genPseudoLegalMoves(Board board) {
		List<Move> moves = new ArrayList<Move>();
		
		Set<Square> squares = Square.allSquares();
		for (Square square : squares) {
			Piece p = board.getPiece(square);

			if (p != null && p.getColor().equals(board.getPlayerToMove())) { 
				if (p instanceof Pawn) {
					genPawnMoves(board,square,moves);
				} else if (p instanceof Knight) {
					genKnightMoves(board,square,moves);				
				} else if (p instanceof Bishop) {
					genBishopMoves(board,square,moves);
				} else if (p instanceof Rook) {
					genRookMoves(board,square,moves);
				} else if (p instanceof Queen) {
					genQueenMoves(board,square,moves);
				} else if (p instanceof King) {
					genKingMoves(board,square,moves);
				}
			}
		}
		
		return moves;
	}
	
	private static void genQueenMoves(Board board,Square fromSq,List<Move> moves) {
		genSliderMoves(PieceDirections.getQueenDirections(),board,fromSq,moves);
	}

	private static void genRookMoves(Board board,Square fromSq,List<Move> moves) {
		genSliderMoves(PieceDirections.getRookDirections(),board,fromSq,moves);
	}

	private static void genSliderMoves(Direction direction,Board board,Square fromSq,List<Move> moves) {
		Color player = board.getPiece(fromSq).getColor();
		Color opponent = Color.swap(player);
		
		Square toSq = direction.next(fromSq);
		while (toSq != null) {
			Piece p = board.getPiece(toSq);
			if (p==null || p.getColor().equals(opponent)) 
				moves.add(new Move(fromSq,toSq,p));
			if (p!=null)
				break;
			toSq = direction.next(toSq);
		}
	}

	private static void genSliderMoves(Set<Direction> directions,Board board,Square fromSq,List<Move> moves) {
		for (Direction dir : directions) {
			genSliderMoves(dir,board,fromSq,moves);
		}
	}

	private static void genWhitePawnAttacks(Direction dir,Board board,Square fromSq,List<Move> moves) {
		Square toSq = dir.next(fromSq);
		if (toSq != null) {
			Piece p = board.getPiece(toSq);
			if (p != null && p.getColor().isBlack()) {
				if (toSq.rank().equals(Rank.RANK_8)) {
					moves.add(new Move(fromSq,toSq,p,Queen.WHITE_QUEEN));
					moves.add(new Move(fromSq,toSq,p,Rook.WHITE_ROOK));
					moves.add(new Move(fromSq,toSq,p,Bishop.WHITE_BISHOP));
					moves.add(new Move(fromSq,toSq,p,Knight.WHITE_KNIGHT));
				} else {
					moves.add(new Move(fromSq,toSq,p));
				}
			} else if (toSq.equals(board.getEPSquare())) {
				Square s = Square.valueOf(board.getEPSquare().file(),board.getEPSquare().rank().south());
				moves.add(new Move(fromSq,toSq,board.getPiece(s)));
			}
		}
	}
	
	private static void genWhitePawnPushes(Board board,Square fromSq,List<Move> moves) {
		Square toSq = North.getInstance().next(fromSq);
		if (board.isEmpty(toSq)) {
			if (toSq.rank().equals(Rank.RANK_8)) {
				moves.add(new Move(fromSq,toSq,null,Queen.WHITE_QUEEN));
				moves.add(new Move(fromSq,toSq,null,Rook.WHITE_ROOK));
				moves.add(new Move(fromSq,toSq,null,Bishop.WHITE_BISHOP));
				moves.add(new Move(fromSq,toSq,null,Knight.WHITE_KNIGHT));
			} else {
				moves.add(new Move(fromSq,toSq,null));
				// double push
				if (fromSq.rank().equals(Rank.RANK_2)) {
					toSq = North.getInstance().next(toSq);
					if (board.isEmpty(toSq))
						moves.add(new Move(fromSq,toSq,null));
				}
			}
		}
	}
	
}
