package chess4j.io;

import java.text.ParseException;

import chess4j.Color;
import chess4j.board.Board;
import chess4j.board.CastlingRights;
import chess4j.board.File;
import chess4j.board.Rank;
import chess4j.board.Square;
import chess4j.pieces.Bishop;
import chess4j.pieces.King;
import chess4j.pieces.Knight;
import chess4j.pieces.Pawn;
import chess4j.pieces.Queen;
import chess4j.pieces.Rook;
import chess4j.utils.BlankRemover;


public class FenParser {
	
	public static void setPos(Board b,String fen) throws ParseException {
		fen = fen.trim();
		b.clearBoard();

		fen = BlankRemover.trim(fen);
		// split on spaces
		String[] fen_pieces = fen.split(" ");
		if (fen_pieces.length < 4) {
			throw new ParseException("not enough parts to FEN.",0);
		}
		
		setPieces(b,fen_pieces[0]);
		setPlayer(b,fen_pieces[1]);
		setCastlingRights(b,fen_pieces[2]);
		setEP(b,fen_pieces[3]);
		
		// Parts 5 and 6 are the half move clock and the full move counter, respectively.
		// They are not part of the EPD specification and are considered optional here.
		
		String halfMoveClock = fen_pieces.length > 4 ? fen_pieces[4] : null;
		setHalfMoveClock(b,halfMoveClock);
		
		String fullMoveCounter = fen_pieces.length > 5 ? fen_pieces[5] : null;
		setFullMoveCounter(b,fullMoveCounter);
	}

	private static void setHalfMoveClock(Board b,String s) throws ParseException {
		try {
			Integer halfMoves = s==null ? 0 : Integer.valueOf(s);
			b.setFiftyCounter(halfMoves);
		} catch (NumberFormatException e) {
			throw new ParseException(e.getMessage(),0);
		}
	}
	
	private static void setFullMoveCounter(Board b,String s) throws ParseException {
		try {
			Integer moveCounter = s==null? 0 : Integer.valueOf(s);
			b.setMoveCounter((moveCounter-1)*2);
			if (Color.BLACK.equals(b.getPlayerToMove())) {
				b.setMoveCounter(b.getMoveCounter()+1);
			}
		} catch (NumberFormatException e) {
			throw new ParseException(e.getMessage(),0);
		}
	}
	
	private static void setCastlingRights(Board b,String s) throws ParseException {
		if (s.equals("-")) {
			return;
		}
		
		char[] arr = s.toCharArray();
		for (int i=0;i<arr.length;i++) {
			switch (arr[i]) {
			case 'K':
				b.addCastlingRight(CastlingRights.WHITE_KINGSIDE);
				break;
			case 'k':
				b.addCastlingRight(CastlingRights.BLACK_KINGSIDE);
				break;
			case 'Q':
				b.addCastlingRight(CastlingRights.WHITE_QUEENSIDE);
				break;
			case 'q':
				b.addCastlingRight(CastlingRights.BLACK_QUEENSIDE);
				break;
			default:
				throw new ParseException("invalid character in setCastlingRights: " + s,i);
			}
		}
	}

	private static void setEP(Board b,String s) throws ParseException {
		if (s.equals("-")) {
			return;
		}
		
		char[] arr = s.toCharArray();
		if (arr.length != 2)
			throw new ParseException("invalid string in setEP: " + s,0);
		
		int epsq=0;
		if (arr[0]>='a' && arr[0]<='h')
			epsq=arr[0]-'a';
		else
			throw new ParseException("invalid string in setEP: " + s,0);
		
		if (arr[1]>='1' && arr[1]<='8')
			epsq+= 8 * (8-(arr[1]-'0'));
		else
			throw new ParseException("invalid string in setEP: " + s,0);
		
		File f = File.file(epsq&7);
		Rank r = Rank.rank(epsq/8);
		b.setEP(Square.valueOf(f, r));
	}

	private static void setPieces(Board b,String s) {
		char[] arr = s.toCharArray();
		int sq=0;
		for (int i=0;i<arr.length;i++) {
			// map sq to a rank and file
			File f = File.file(sq&7);
			Rank r = Rank.rank(sq/8);
			Square square = Square.valueOf(f, r);
			switch (arr[i]) {
			case 'K':
				b.addPiece(King.WHITE_KING,square);
				sq++;
				break;
			case 'k':
				b.addPiece(King.BLACK_KING,square);
				sq++;
				break;
			case 'Q':
				b.addPiece(Queen.WHITE_QUEEN,square);
				sq++;
				break;
			case 'q':
				b.addPiece(Queen.BLACK_QUEEN,square);
				sq++;
				break;
			case 'R':
				b.addPiece(Rook.WHITE_ROOK,square);
				sq++;
				break;
			case 'r':
				b.addPiece(Rook.BLACK_ROOK,square);
				sq++;
				break;
			case 'N':
				b.addPiece(Knight.WHITE_KNIGHT,square);
				sq++;
				break;
			case 'n':
				b.addPiece(Knight.BLACK_KNIGHT,square);
				sq++;
				break;
			case 'B':
				b.addPiece(Bishop.WHITE_BISHOP,square);
				sq++;
				break;
			case 'b':
				b.addPiece(Bishop.BLACK_BISHOP,square);
				sq++;
				break;
			case 'P':
				b.addPiece(Pawn.WHITE_PAWN,square);
				sq++;
				break;
			case 'p':
				b.addPiece(Pawn.BLACK_PAWN,square);
				sq++;
				break;
			case '8': sq+=8; break;
			case '7': sq+=7; break;
			case '6': sq+=6; break;
			case '5': sq+=5; break;
			case '4': sq+=4; break;
			case '3': sq+=3; break;
			case '2': sq+=2; break;
			case '1': sq++;  break;
			}
		}
		assert(sq==64);
	}

	private static void setPlayer(Board b,String s) throws ParseException {
		Color ptm;
		if (s.equalsIgnoreCase("b"))
			ptm = Color.BLACK;
		else if (s.equalsIgnoreCase("w"))
			ptm = Color.WHITE;
		else
			throw new ParseException("could not parse player: " + s,0);
		
		if (!ptm.equals(b.getPlayerToMove()))
			b.swapPlayer();
	}

}
