package chess4j.search;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import chess4j.Constants;
import chess4j.board.Board;
import chess4j.eval.Eval;
import chess4j.io.PrintLine;
import chess4j.moves.Move;
import chess4j.moves.MoveGen;
import chess4j.utils.MoveUtils;

public class Search {
	
	private static final Log logger = LogFactory.getLog(Search.class);

	private static boolean abortSearch=false;
	
	private Search() {}

	public static int search(List<Move> parent_pv,int alpha,int beta,Board board,
			int depth,SearchStats stats) {
		
		assert(depth>0);
		assert(beta>alpha);
		
		stats.getFirstLine().clear();
		stats.incNodes();
		
		List<Move> pv = new ArrayList<Move>();
		List<Move> moves = MoveGen.genLegalMoves(board);
		// place previous PV move on top
		if (stats.getLastPV().size()>0) {
			MoveUtils.putMoveAtTop(moves, stats.getLastPV().get(0));
		}
		
		int numMovesSearched=0;
		for (Move move : moves) {
			board.applyMove(move);
			if (numMovesSearched==0) {
				stats.getFirstLine().add(move);
			}
			int score = -searchHelper(pv,-beta,-alpha,board,depth-1,1,(numMovesSearched==0),stats);
			numMovesSearched++;
			board.undoLastMove();
			
			// if depth==1, we need to continue to search to ensure we have something to play
			if (depth > 1 && abortSearch) {
				return 0;
			}
			
			if (score > alpha) {
				if (score >= beta) {
					return beta;
				}
				alpha = score;
				// update the pv
				parent_pv.clear();
				parent_pv.add(move);
				parent_pv.addAll(pv);
				// print thinking output
				PrintLine.printLine(parent_pv,depth,score,stats.getStartTime(),stats.getNodes());
			}
		}
		
		// is this a mate?
		if (numMovesSearched==0) {
			if (board.isPlayerInCheck()) {
				// adjust for distance from root
				return -Constants.CHECKMATE; 
			} else {
				return 0; // draw score
			}
		}
		
		return alpha;
	}
	
	private static int searchHelper(List<Move> parent_pv,int alpha,int beta,Board board,
			int depth,int ply,boolean pvNode,SearchStats stats) {

		assert(ply > 0);
		assert(alpha < beta);
		
		parent_pv.clear();
		stats.incNodes();
		if (depth < 1) {
			return Eval.eval(board);
		}

		// TODO: don't check EVERY node for time
		if (System.currentTimeMillis() > stats.getStopTime()) {
			abortSearch = true;
			logger.debug("# Aborting search depth " + depth + ", ply=" + ply + " on time...");
			return 0;
		}
		
		List<Move> pv = new ArrayList<Move>();
		List<Move> moves = MoveGen.genLegalMoves(board);
		// if pv node put previous PV move on top
		if (pvNode && stats.getLastPV().size() > ply) {
			MoveUtils.putMoveAtTop(moves, stats.getLastPV().get(ply));
		}
		
		int numMovesSearched=0;
		for (Move move : moves) {
			board.applyMove(move);
			if (pvNode && numMovesSearched==0) {
				stats.getFirstLine().add(move);
			}
			int score = -searchHelper(pv,-beta,-alpha,board,depth-1,ply+1,
					(pvNode && numMovesSearched==0),stats);
			numMovesSearched++;
			board.undoLastMove();
			
			if (abortSearch) {
				return 0;
			}
			
			if (score > alpha) {
				if (score >= beta) {
					return beta;
				}
				alpha = score;
				// update the pv
				parent_pv.clear();
				parent_pv.add(move);
				parent_pv.addAll(pv);
			}
		}
		
		// is this a mate?
		if (numMovesSearched==0) {
			if (board.isPlayerInCheck()) {
				// adjust for distance from root
				return -(Constants.CHECKMATE-ply); 
			} else {
				return 0; // draw score
			}
		}

		return alpha;
	}


	public static void setAbortSearch(boolean abortSearch) {
		Search.abortSearch = abortSearch;
	}

	public static boolean isAbortSearch() {
		return abortSearch;
	}
	
}
